#ifndef ANIMATION_H
#define ANIMATION_H
#include "maths.h"

struct frame {
  unsigned int offset;
  vector_t position; /* relative to origin object */
  vector_t root_position;
  quaternion_t q;
  unsigned short num_events;
  unsigned char **events;
  unsigned char changed;
};

struct event_set {
  struct event_set *next;
  unsigned short frame_id;
  unsigned char *event;
};

struct rotation_set {
  struct rotation_set *next;
  char *bone_arg;
  unsigned short bone_id;
  unsigned short first_frame;
  int last_frame;
  matrix_t rotation;
};

GSM_API int find_animation_stuff(struct gsm *);
GSM_API struct frame **alloc_frames(unsigned short);
GSM_API struct frame *alloc_frame();
GSM_API void cleanup_triangles(unsigned short, struct triangle **);
GSM_API int find_frames(struct gsm *);
GSM_API int parse_frame(struct gsm *, unsigned short, struct frame *, unsigned short);
GSM_API int find_events(struct gsm *, struct frame *, unsigned short);
GSM_API int write_animation_stuff(struct gsm *);
GSM_API int write_frames(struct gsm *);
GSM_API int write_frame(struct gsm *, unsigned short, struct frame *, unsigned short);
GSM_API int write_events(struct gsm *, struct frame *, unsigned short);
GSM_API int delete_bone_from_frames(struct gsm *, unsigned short);
GSM_API unsigned char **alloc_events(unsigned short);
GSM_API unsigned char *alloc_event(unsigned short);
GSM_API void cleanup_events(unsigned short, unsigned char **);
GSM_API struct event_set *alloc_event_set();
GSM_API int add_to_event_set(struct event_set **, char *, char *);
GSM_API int set_events(struct gsm *);
GSM_API char *get_events(struct frame *);
GSM_API struct rotation_set *alloc_rotation_set();
GSM_API void cleanup_rotation_sets(struct rotation_set *);
GSM_API int add_to_rotation_set(struct rotation_set **, char *, char *, char *, char *, char *);
GSM_API int rotate_bones(struct gsm *);
GSM_API int rotate_bone(struct gsm *, struct rotation_set *);
GSM_API int calculate_world_for_all_children(struct gsm *, unsigned short, unsigned short);
GSM_API int calculate_animation_matrices(struct gsm *, unsigned short);
#endif
