#include "gsmhacking.h"

struct assignment **alloc_assignments(unsigned short n) {
  return (struct assignment **) alloc_lots(n, "assignment", sizeof(struct assignment *), (void *(*)()) alloc_assignment);
}

struct assignment *alloc_assignment() {
  return (struct assignment *) alloc_one("assignment", sizeof(struct assignment));
}

void cleanup_assignments(unsigned short num_assignments, struct assignment **assignments) {
  free_structs("assignment", num_assignments, (void **) assignments);
}

int find_assignments(struct gsm *gsm) {
  int i;
  unsigned int pos;
  struct assignment *assignment;

  if (! gsm || ! gsm->fd) {
    ERROR("no file descriptor for gsm!\n");
    return 1;
  }

  pos = get_file_position(gsm->fd);
  TRACE("get_file_position({gsm->fd=%p}: 0x%04x\n", gsm->fd, pos);
  DEBUG("we think assignments start at 0x%04x\n", pos);

  /* how many assignments are there? */
  if (read_short(gsm->fd, &gsm->num_assignments)) {
    ERROR("can't read number of assignments");
    return 2;
  }
  if (! gsm->num_assignments) {
    SAY("%s has no assignments\n", gsm->filename);
    return 3;
  }
  DEBUG("vertex to bone assignments: %u\n", gsm->num_assignments);

  /* allocate a big enough struct */
  gsm->assignments = alloc_assignments(gsm->num_assignments);
  if (! gsm->assignments) return 4;

  /* fill in the assignments */
  for (i = 0; i < gsm->num_assignments; i++) {
    assignment = gsm->assignments[i];
    pos = get_file_position(gsm->fd);
    if (pos == -1) return 5;
    assignment->bone_id = -1;
    assignment->offset = pos;
    if (parse_assignment(gsm, i, assignment, gsm->num_assignments)) {
      ERROR("expected assignment id %u but got null\n", i);
      return 6;
    }
  }

  return 0;
}

/* set bone explicitly for positions - will need for Milkshape later */
int assign_bones(struct gsm *gsm) {
  unsigned short i, j, k;
  struct assignment *assignment;

  if (! gsm) {
    ERROR("no GSM!\n");
    return 1;
  }

  DEBUG("setting bone IDs from %u assignments\n", gsm->num_assignments);
  for (i = 0; i < gsm->num_assignments; i++) {
    assignment = gsm->assignments[i];
    TRACE("setting bone ID for %u positions from %u\n", assignment->num_vertices, assignment->first_vertex);
    for (j = 0; j < assignment->num_vertices; j++) {
      k = j + assignment->first_vertex; 
      if (k >= gsm->num_positions) {
        ERROR("position %u out of range\n", k);
        return 3;
      }
      gsm->positions[k]->bone_id = assignment->bone_id;
      TRACE1("assigning bone %u to position %u\n", assignment->bone_id, k);
    }
  }

  return 0;
}

int parse_assignment(struct gsm *gsm, unsigned short num, struct assignment *assignment, unsigned short total) {
  int i, skipped, unordered;

  /* find id */
  if (read_short(gsm->fd, &assignment->bone_id)) return 1;

  /*
    we're going to assume that one of two possibilities is true

      1) there are the expected number of assignments

      2) there are fewer than expected but there's a magic 0xee 0xee in 
         place of the last assignment number
  */
  if (assignment->bone_id == assignments_hack) {
    DEBUG("we hit end of assignments list\n");
    gsm->need_assignments_hack = 1;
    /* count up assignment id in gsm */
    /* FIXME: this code serves no useful purpose other than debugging now */
    unordered = 0;
    for (i = 0; i < num - 1; i++) {
      if (gsm->assignments[i]->bone_id != i + unordered) {
        TRACE("gsm->assignments[%u]->bone_id = %u (expected %u)\n", i, gsm->assignments[i]->bone_id, i + unordered);
        unordered++;
      }
    }

    /*
       FIXME FIXME FIXME
       scrap that bone_id = -1 hack!
    */

    skipped = total - num;
    if (skipped) {
      TRACE("looks like some assignments were missing (%u to be precise)\n", skipped);
    }
    else if (assignment->bone_id > total + 1) {
      DEBUG("no assignments were missing up to now");
    }

    /* free unneeded assignments */
    /* FIXME: this is fubar */
    for (i = total - 1; i > total - 1 - skipped; i--) {
      TRACE("freeing unnecessary assignment #%u at %p\n", i, gsm->assignments[i]);
      free(gsm->assignments[i]);
    }
    gsm->num_assignments -= skipped;
    TRACE("set correct number of assignments back %d to %u\n", skipped, gsm->num_assignments);

    return 0;
  }
  else if (assignment->bone_id > total) {
    TRACE("unexpectedly high bone id 0x%04x for assignment %u\n", assignment->bone_id, num);
  }

  /* find first vertex in the assignment */
  if (read_short(gsm->fd, &assignment->first_vertex)) return 2;
  if (read_short(gsm->fd, &assignment->num_vertices)) return 3;
  if (read_short(gsm->fd, &assignment->unknown)) return 3;

  TRACE("<0x%04x>: <0x%04x> <0x%04x> <0x%04x> <0x%04x>\n", assignment->offset, assignment->bone_id, assignment->first_vertex, assignment->num_vertices, assignment->unknown);
  DEBUG("assignment %u at 0x%04x: %02d %s: %u vertices from %u to %u\n", num, assignment->offset, assignment->bone_id, get_bone_by_id(gsm, assignment->bone_id), assignment->num_vertices, assignment->first_vertex, assignment->first_vertex + assignment->num_vertices - 1);

  return 0;
}

int write_assignments(struct gsm *gsm) {
  unsigned short i;
  unsigned short num_assignments;

  if (! gsm || ! gsm->out_fd) {
    ERROR("no output file descriptor for gsm!\n");
    return 1;
  }

  num_assignments = gsm->num_assignments + gsm->need_assignments_hack;
  if (gsm->need_assignments_hack) {
    TRACE("assignments hack is needed, claiming there are %u assignments\n", num_assignments);
  }
  DEBUG("writing assignments: %u\n", num_assignments);
  if (write_short(gsm->out_fd, &num_assignments)) {
    ERROR("can't write number of assignments");
    return 2;
  }

  /* only write out the real number */
  for (i = 0; i < gsm->num_assignments; i++) {
    TRACE("writing assignment %u\n", i);
    if (write_short(gsm->out_fd, &gsm->assignments[i]->bone_id)) {
      ERROR("can't write assignment %u bone id", i);
      return 3;
    }
    if (write_short(gsm->out_fd, &gsm->assignments[i]->first_vertex)) {
      ERROR("can't write assignment %u first vertex", i);
      return 4;
    }
    if (write_short(gsm->out_fd, &gsm->assignments[i]->num_vertices)) {
      ERROR("can't write assignment %u num vertices", i);
      return 5;
    }
    if (write_short(gsm->out_fd, &gsm->assignments[i]->unknown)) {
      ERROR("can't write assignment %u unknown", i);
      return 6;
    }
  }

  /* write out the magic 0xEE 0xEE */
  if (gsm->need_assignments_hack) {
    TRACE("writing 0xEE 0xEE to terminate assignments\n");
    if (write_short(gsm->out_fd, &assignments_hack)) {
      ERROR("can't write assignments hack\n");
      return 7;
    }
  }

  return 0;
}

/* renumber bones in assignments */
int reassign_kid_to_swt(struct gsm *gsm) {
  unsigned short i, j, old;
  int new, gripper_id;

  if (gsm->num_assignments == 1) {
    TRACE("converting one assignment\n");
  }
  else TRACE("converting %u assignments\n", gsm->num_assignments);

  for (i = 0; i < gsm->num_assignments; i++) {
    old = gsm->assignments[i]->bone_id;
    switch (old) {
      /* kids have 6 ponytails */
      case KID_BIPED_PONYTAIL1:
      case KID_BIPED_PONYTAIL11:
      case KID_BIPED_PONYTAIL12:
      case KID_BIPED_PONYTAIL2:
      case KID_BIPED_PONYTAIL21:
      case KID_BIPED_PONYTAIL22:
        /* link vertices to last valid bone */
        if (gsm->assignments[i]->num_vertices) {
          gsm->assignments[i - 1]->num_vertices += gsm->assignments[i]->num_vertices;
          TRACE("old bone %u was %s: vertices moved to %s: %u\n", i, get_biped_bone_name(GSM_TYPE_KID, old), get_biped_bone_name(GSM_TYPE_SWT, gsm->assignments[i - 1]->bone_id), gsm->assignments[i]->num_vertices);
        }
        /* free this unnecessary assignment */
        TRACE("freeing unneeded assignment %u\n", i);
        free(gsm->assignments[i]);
        gsm->num_assignments--;
        for (j = i; j < gsm->num_assignments; j++) {
          gsm->assignments[j] = gsm->assignments[j + 1];
        }
        /* fall through to next case to deal with this one */
        i--;
        continue;
      case KID_BIPED_PELVIS:
      case KID_BIPED_SPINE:
      case KID_BIPED_SPINE1:
      case KID_BIPED_SPINE2:
      case KID_BIPED_NECK:
      case KID_BIPED_HEAD:
      case KID_BIPED_L_CLAVICLE:
      case KID_BIPED_L_UPPERARM:
      case KID_BIPED_L_FOREARM:
      case KID_BIPED_L_HAND:
      case KID_BIPED_L_FINGER0:
      case KID_BIPED_L_FINGER01:
      case KID_BIPED_L_FINGER1:
      case KID_BIPED_L_FINGER11:
      case KID_BIPED_L_FINGER2:
      case KID_BIPED_L_FINGER21:
      case KID_BIPED_R_CLAVICLE:
      case KID_BIPED_R_UPPERARM:
      case KID_BIPED_R_FOREARM:
      case KID_BIPED_R_HAND:
      case KID_BIPED_R_FINGER0:
      case KID_BIPED_R_FINGER01:
      case KID_BIPED_R_FINGER1:
      case KID_BIPED_R_FINGER11:
      case KID_BIPED_R_FINGER2:
      case KID_BIPED_R_FINGER21:
      case KID_BIPED_L_THIGH:
      case KID_BIPED_L_CALF:
      case KID_BIPED_L_FOOT:
      case KID_BIPED_L_TOE0:
      case KID_BIPED_R_THIGH:
      case KID_BIPED_R_CALF:
      case KID_BIPED_R_FOOT:
      case KID_BIPED_R_TOE0:
        new = reassign_bone_id_kid_to_swt(old);
        if (new < 0 || new == BONE_FIRST_LINK) {
          ERROR("bogus bone reassignment from %u to %d\n", old, new);
          return i + 1;
        }
        gsm->assignments[i]->bone_id = new;
        break;
      default:
        ERROR("WARNING: unexpected bone_id %u in assignment %u\n", gsm->assignments[i]->bone_id, i);
        return i + 1;
    }
    TRACE1("assignment %u converted from %02d %s to %02d %s\n", i, old, get_bone_by_id(gsm, old), gsm->assignments[i]->bone_id, get_biped_bone_name(GSM_TYPE_SWT, gsm->assignments[i]->bone_id));
    gripper_id = get_gripper_by_bone(gsm, old);
    if (gripper_id > -1) gsm->grippers[gripper_id]->bone_id = gsm->assignments[i]->bone_id;
  }

  return 0;
}

/* renumber bones in assignments */
int reassign_tac_to_swt(struct gsm *gsm) {
  unsigned short i, j, old;
  int new, gripper_id;

  if (gsm->num_assignments == 1) {
    TRACE("converting one assignment\n");
  }
  else TRACE("converting %u assignments\n", gsm->num_assignments);

  for (i = 0; i < gsm->num_assignments; i++) {
    old = gsm->assignments[i]->bone_id;
    switch (old) {
      case TAC_BIPED_PONYTAIL1:
      case TAC_BIPED_PONYTAIL11:
      case TAC_BIPED_PONYTAIL12:
        /* link vertices to last valid bone */
        if (gsm->assignments[i]->num_vertices) {
          gsm->assignments[i - 1]->num_vertices += gsm->assignments[i]->num_vertices;
          TRACE("old bone %u was %s: vertices moved to %s: %u\n", i, get_biped_bone_name(GSM_TYPE_TAC, old), get_biped_bone_name(GSM_TYPE_SWT, gsm->assignments[i - 1]->bone_id), gsm->assignments[i]->num_vertices);
        }
        /* free this unnecessary assignment */
        TRACE("freeing unneeded assignment %u\n", i);
        free(gsm->assignments[i]);
        gsm->num_assignments--;
        for (j = i; j < gsm->num_assignments; j++) {
          gsm->assignments[j] = gsm->assignments[j + 1];
        }
        /* fall through to next case to deal with this one */
        i--;
        continue;
      case TAC_BIPED_PELVIS:
      case TAC_BIPED_SPINE:
      case TAC_BIPED_SPINE1:
      case TAC_BIPED_SPINE2:
      case TAC_BIPED_NECK:
      case TAC_BIPED_HEAD:
      case TAC_BIPED_L_CLAVICLE:
      case TAC_BIPED_L_UPPERARM:
      case TAC_BIPED_L_FOREARM:
      case TAC_BIPED_L_HAND:
      case TAC_BIPED_L_FINGER0:
      case TAC_BIPED_L_FINGER01:
      case TAC_BIPED_L_FINGER1:
      case TAC_BIPED_L_FINGER11:
      case TAC_BIPED_L_FINGER2:
      case TAC_BIPED_L_FINGER21:
      case TAC_BIPED_R_CLAVICLE:
      case TAC_BIPED_R_UPPERARM:
      case TAC_BIPED_R_FOREARM:
      case TAC_BIPED_R_HAND:
      case TAC_BIPED_R_FINGER0:
      case TAC_BIPED_R_FINGER01:
      case TAC_BIPED_R_FINGER1:
      case TAC_BIPED_R_FINGER11:
      case TAC_BIPED_R_FINGER2:
      case TAC_BIPED_R_FINGER21:
      case TAC_BIPED_L_THIGH:
      case TAC_BIPED_L_CALF:
      case TAC_BIPED_L_FOOT:
      case TAC_BIPED_L_TOE0:
      case TAC_BIPED_R_THIGH:
      case TAC_BIPED_R_CALF:
      case TAC_BIPED_R_FOOT:
      case TAC_BIPED_R_TOE0:
        new = reassign_bone_id_tac_to_swt(old);
        if (new < 0 || new == BONE_FIRST_LINK) {
          ERROR("bogus bone reassignment from %u to %d\n", old, new);
          return i + 1;
        }
        gsm->assignments[i]->bone_id = new;
        break;
      default:
        ERROR("WARNING: unexpected bone_id %u in assignment %u\n", gsm->assignments[i]->bone_id, i);
        return i + 1;
    }
    TRACE1("assignment %u converted from %02d %s to %02d %s\n", i, old, get_bone_by_id(gsm, old), gsm->assignments[i]->bone_id, get_biped_bone_name(GSM_TYPE_SWT, gsm->assignments[i]->bone_id));
    gripper_id = get_gripper_by_bone(gsm, old);
    if (gripper_id > -1) gsm->grippers[gripper_id]->bone_id = gsm->assignments[i]->bone_id;
  }

  return 0;
}

/* renumber bones in assignments */
int reassign_bf1_to_swt(struct gsm *gsm) {
  unsigned short i, old;
  int new, gripper_id;

  if (gsm->num_assignments == 1) {
    TRACE("converting one assignment\n");
  }
  else TRACE("converting %d assignments\n", gsm->num_assignments);

  for (i = 0; i < gsm->num_assignments; i++) {
    old = gsm->assignments[i]->bone_id;
    new = reassign_bone_id_bf1_to_swt(old);
    if (new < 0 || new == BONE_FIRST_LINK) {
      ERROR("bogus bone reassignment from %u to %d\n", old, new);
      return i + 1;
    }
    gsm->assignments[i]->bone_id = new;
    TRACE1("assignment %u converted from %02u %s to %02u %s\n", i, old, get_bone_by_id(gsm, old), gsm->assignments[i]->bone_id, get_biped_bone_name(GSM_TYPE_SWT, gsm->assignments[i]->bone_id));
    gripper_id = get_gripper_by_bone(gsm, old);
    if (gripper_id > -1) gsm->grippers[gripper_id]->bone_id = gsm->assignments[i]->bone_id;
  }

  return 0;
}

/* renumber bones in assignments */
int reassign_swt_to_kid(struct gsm *gsm) {
  unsigned short i, j, old;
  int new, gripper_id;

  if (gsm->num_assignments == 1) {
    TRACE("converting one assignment\n");
  }
  else TRACE("converting %u assignments\n", gsm->num_assignments);

  for (i = 0; i < gsm->num_assignments; i++) {
    old = gsm->assignments[i]->bone_id;
    switch (old) {
      /* extra fingers unique to SWT */
      case SWT_BIPED_L_FINGER02:
      case SWT_BIPED_L_FINGER12:
      case SWT_BIPED_L_FINGER22:
      case SWT_BIPED_L_FINGER3:
      case SWT_BIPED_L_FINGER31:
      case SWT_BIPED_L_FINGER32:
      case SWT_BIPED_L_FINGER4:
      case SWT_BIPED_L_FINGER41:
      case SWT_BIPED_L_FINGER42:
      case SWT_BIPED_R_FINGER02:
      case SWT_BIPED_R_FINGER12:
      case SWT_BIPED_R_FINGER22:
      case SWT_BIPED_R_FINGER3:
      case SWT_BIPED_R_FINGER31:
      case SWT_BIPED_R_FINGER32:
      case SWT_BIPED_R_FINGER4:
      case SWT_BIPED_R_FINGER41:
      case SWT_BIPED_R_FINGER42:
        /* link vertices to last valid bone */
        if (gsm->assignments[i]->num_vertices) {
          gsm->assignments[i - 1]->num_vertices += gsm->assignments[i]->num_vertices;
          TRACE("old bone %u was %s: vertices moved to %s: %u\n", i, get_biped_bone_name(GSM_TYPE_SWT, old), get_biped_bone_name(GSM_TYPE_KID, gsm->assignments[i - 1]->bone_id), gsm->assignments[i]->num_vertices);
        }
        /* free this unnecessary assignment */
        TRACE("freeing unneeded assignment %u\n", i);
        free(gsm->assignments[i]);
        gsm->num_assignments--;
        for (j = i; j < gsm->num_assignments; j++) {
          gsm->assignments[j] = gsm->assignments[j + 1];
        }
        /* fall through to next case to deal with this one */
        i--;
        continue;
      case SWT_BIPED_PELVIS:
      case SWT_BIPED_SPINE:
      case SWT_BIPED_SPINE1:
      case SWT_BIPED_SPINE2:
      case SWT_BIPED_NECK:
      case SWT_BIPED_HEAD:
      case SWT_BIPED_L_CLAVICLE:
      case SWT_BIPED_L_UPPERARM:
      case SWT_BIPED_L_FOREARM:
      case SWT_BIPED_L_HAND:
      case SWT_BIPED_L_FINGER0:
      case SWT_BIPED_L_FINGER01:
      case SWT_BIPED_L_FINGER1:
      case SWT_BIPED_L_FINGER11:
      case SWT_BIPED_L_FINGER2:
      case SWT_BIPED_L_FINGER21:
      case SWT_BIPED_R_CLAVICLE:
      case SWT_BIPED_R_UPPERARM:
      case SWT_BIPED_R_FOREARM:
      case SWT_BIPED_R_HAND:
      case SWT_BIPED_R_FINGER0:
      case SWT_BIPED_R_FINGER01:
      case SWT_BIPED_R_FINGER1:
      case SWT_BIPED_R_FINGER11:
      case SWT_BIPED_R_FINGER2:
      case SWT_BIPED_R_FINGER21:
      case SWT_BIPED_L_THIGH:
      case SWT_BIPED_L_CALF:
      case SWT_BIPED_L_FOOT:
      case SWT_BIPED_L_TOE0:
      case SWT_BIPED_R_THIGH:
      case SWT_BIPED_R_CALF:
      case SWT_BIPED_R_FOOT:
      case SWT_BIPED_R_TOE0:
        new = reassign_bone_id_swt_to_kid(old);
        if (new < 0 || new == BONE_FIRST_LINK) {
          ERROR("bogus bone reassignment from %u to %d\n", old, new);
          return i + 1;
        }
        gsm->assignments[i]->bone_id = new;
        break;
      default:
        ERROR("WARNING: unexpected bone_id %u in assignment %u\n", gsm->assignments[i]->bone_id, i);
        return i + 1;
    }
    TRACE1("assignment %u converted from %02u %s to %02u %s\n", i, old, get_bone_by_id(gsm, old), gsm->assignments[i]->bone_id, get_biped_bone_name(GSM_TYPE_KID, gsm->assignments[i]->bone_id));
  }

  return 0;
}

/* renumber bones in assignments */
int reassign_tac_to_kid(struct gsm *gsm) {
  unsigned short i, old;
  int new, gripper_id;

  if (gsm->num_assignments == 1) {
    TRACE("converting one assignment\n");
  }
  else TRACE("converting %d assignments\n", gsm->num_assignments);

  for (i = 0; i < gsm->num_assignments; i++) {
    old = gsm->assignments[i]->bone_id;
    new = reassign_bone_id_tac_to_kid(old);
    if (new < 0 || new == BONE_FIRST_LINK) {
      ERROR("bogus bone reassignment from %u to %d\n", old, new);
      return i + 1;
    }
    gsm->assignments[i]->bone_id = new;
    TRACE1("assignment %u converted from %02u %s to %02u %s\n", i, old, get_bone_by_id(gsm, old), gsm->assignments[i]->bone_id, get_biped_bone_name(GSM_TYPE_KID, gsm->assignments[i]->bone_id));
  }

  return 0;
}

/* renumber bones in assignments */
int reassign_bf1_to_kid(struct gsm *gsm) {
  unsigned short i, old;
  int new, gripper_id;

  if (gsm->num_assignments == 1) {
    TRACE("converting one assignment\n");
  }
  else TRACE("converting %d assignments\n", gsm->num_assignments);

  for (i = 0; i < gsm->num_assignments; i++) {
    old = gsm->assignments[i]->bone_id;
    new = reassign_bone_id_bf1_to_kid(old);
    if (new < 0 || new == BONE_FIRST_LINK) {
      ERROR("bogus bone reassignment from %u to %d\n", old, new);
      return i + 1;
    }
    gsm->assignments[i]->bone_id = new;
    TRACE1("assignment %u converted from %02u %s to %02u %s\n", i, old, get_bone_by_id(gsm, old), gsm->assignments[i]->bone_id, get_biped_bone_name(GSM_TYPE_KID, gsm->assignments[i]->bone_id));
  }

  return 0;
}

/* renumber bones in assignments */
int reassign_swt_to_tac(struct gsm *gsm) {
  unsigned short i, j, old;
  int new, gripper_id;

  if (gsm->num_assignments == 1) {
    TRACE("converting one assignment\n");
  }
  else TRACE("converting %u assignments\n", gsm->num_assignments);

  for (i = 0; i < gsm->num_assignments; i++) {
    old = gsm->assignments[i]->bone_id;
    switch (old) {
      /* extra fingers unique to SWT */
      case SWT_BIPED_L_FINGER02:
      case SWT_BIPED_L_FINGER12:
      case SWT_BIPED_L_FINGER22:
      case SWT_BIPED_L_FINGER3:
      case SWT_BIPED_L_FINGER31:
      case SWT_BIPED_L_FINGER32:
      case SWT_BIPED_L_FINGER4:
      case SWT_BIPED_L_FINGER41:
      case SWT_BIPED_L_FINGER42:
      case SWT_BIPED_R_FINGER02:
      case SWT_BIPED_R_FINGER12:
      case SWT_BIPED_R_FINGER22:
      case SWT_BIPED_R_FINGER3:
      case SWT_BIPED_R_FINGER31:
      case SWT_BIPED_R_FINGER32:
      case SWT_BIPED_R_FINGER4:
      case SWT_BIPED_R_FINGER41:
      case SWT_BIPED_R_FINGER42:
        /* link vertices to last valid bone */
        if (gsm->assignments[i]->num_vertices) {
          gsm->assignments[i - 1]->num_vertices += gsm->assignments[i]->num_vertices;
          TRACE("old bone %u was %s: vertices moved to %s: %u\n", i, get_biped_bone_name(GSM_TYPE_SWT, old), get_biped_bone_name(GSM_TYPE_TAC, gsm->assignments[i - 1]->bone_id), gsm->assignments[i]->num_vertices);
        }
        /* free this unnecessary assignment */
        TRACE("freeing unneeded assignment %u\n", i);
        free(gsm->assignments[i]);
        gsm->num_assignments--;
        for (j = i; j < gsm->num_assignments; j++) {
          gsm->assignments[j] = gsm->assignments[j + 1];
        }
        /* fall through to next case to deal with this one */
        i--;
        continue;
      case SWT_BIPED_PELVIS:
      case SWT_BIPED_SPINE:
      case SWT_BIPED_SPINE1:
      case SWT_BIPED_SPINE2:
      case SWT_BIPED_NECK:
      case SWT_BIPED_HEAD:
      case SWT_BIPED_L_CLAVICLE:
      case SWT_BIPED_L_UPPERARM:
      case SWT_BIPED_L_FOREARM:
      case SWT_BIPED_L_HAND:
      case SWT_BIPED_L_FINGER0:
      case SWT_BIPED_L_FINGER01:
      case SWT_BIPED_L_FINGER1:
      case SWT_BIPED_L_FINGER11:
      case SWT_BIPED_L_FINGER2:
      case SWT_BIPED_L_FINGER21:
      case SWT_BIPED_R_CLAVICLE:
      case SWT_BIPED_R_UPPERARM:
      case SWT_BIPED_R_FOREARM:
      case SWT_BIPED_R_HAND:
      case SWT_BIPED_R_FINGER0:
      case SWT_BIPED_R_FINGER01:
      case SWT_BIPED_R_FINGER1:
      case SWT_BIPED_R_FINGER11:
      case SWT_BIPED_R_FINGER2:
      case SWT_BIPED_R_FINGER21:
      case SWT_BIPED_L_THIGH:
      case SWT_BIPED_L_CALF:
      case SWT_BIPED_L_FOOT:
      case SWT_BIPED_L_TOE0:
      case SWT_BIPED_R_THIGH:
      case SWT_BIPED_R_CALF:
      case SWT_BIPED_R_FOOT:
      case SWT_BIPED_R_TOE0:
        new = reassign_bone_id_swt_to_tac(old);
        if (new < 0 || new == BONE_FIRST_LINK) {
          ERROR("bogus bone reassignment from %u to %d\n", old, new);
          return i + 1;
        }
        gsm->assignments[i]->bone_id = new;
        break;
      default:
        ERROR("WARNING: unexpected bone_id %u in assignment %u\n", gsm->assignments[i]->bone_id, i);
        return i + 1;
    }
    TRACE1("assignment %u converted from %02u %s to %02u %s\n", i, old, get_bone_by_id(gsm, old), gsm->assignments[i]->bone_id, get_biped_bone_name(GSM_TYPE_TAC, gsm->assignments[i]->bone_id));
    gripper_id = get_gripper_by_bone(gsm, old);
    if (gripper_id > -1) gsm->grippers[gripper_id]->bone_id = gsm->assignments[i]->bone_id;
  }

  return 0;
}

/* renumber bones in assignments */
int reassign_kid_to_tac(struct gsm *gsm) {
  unsigned short i, j, old;
  int new, gripper_id;

  if (gsm->num_assignments == 1) {
    TRACE("converting one assignment\n");
  }
  else TRACE("converting %u assignments\n", gsm->num_assignments);

  for (i = 0; i < gsm->num_assignments; i++) {
    old = gsm->assignments[i]->bone_id;
    switch (old) {
      /* kids have 6 ponytails */
      case KID_BIPED_PONYTAIL2:
      case KID_BIPED_PONYTAIL21:
      case KID_BIPED_PONYTAIL22:
        /* link vertices to last valid bone */
        if (gsm->assignments[i]->num_vertices) {
          gsm->assignments[i - 1]->num_vertices += gsm->assignments[i]->num_vertices;
          TRACE("old bone %u was %s: vertices moved to %s: %u\n", i, get_biped_bone_name(GSM_TYPE_KID, old), get_biped_bone_name(GSM_TYPE_TAC, gsm->assignments[i - 1]->bone_id), gsm->assignments[i]->num_vertices);
        }
        /* free this unnecessary assignment */
        TRACE("freeing unneeded assignment %u\n", i);
        free(gsm->assignments[i]);
        gsm->num_assignments--;
        for (j = i; j < gsm->num_assignments; j++) {
          gsm->assignments[j] = gsm->assignments[j + 1];
        }
        /* fall through to next case to deal with this one */
        i--;
        continue;
      case KID_BIPED_PELVIS:
      case KID_BIPED_SPINE:
      case KID_BIPED_SPINE1:
      case KID_BIPED_SPINE2:
      case KID_BIPED_NECK:
      case KID_BIPED_HEAD:
      case KID_BIPED_PONYTAIL1:
      case KID_BIPED_PONYTAIL11:
      case KID_BIPED_PONYTAIL12:
      case KID_BIPED_L_CLAVICLE:
      case KID_BIPED_L_UPPERARM:
      case KID_BIPED_L_FOREARM:
      case KID_BIPED_L_HAND:
      case KID_BIPED_L_FINGER0:
      case KID_BIPED_L_FINGER01:
      case KID_BIPED_L_FINGER1:
      case KID_BIPED_L_FINGER11:
      case KID_BIPED_L_FINGER2:
      case KID_BIPED_L_FINGER21:
      case KID_BIPED_R_CLAVICLE:
      case KID_BIPED_R_UPPERARM:
      case KID_BIPED_R_FOREARM:
      case KID_BIPED_R_HAND:
      case KID_BIPED_R_FINGER0:
      case KID_BIPED_R_FINGER01:
      case KID_BIPED_R_FINGER1:
      case KID_BIPED_R_FINGER11:
      case KID_BIPED_R_FINGER2:
      case KID_BIPED_R_FINGER21:
      case KID_BIPED_L_THIGH:
      case KID_BIPED_L_CALF:
      case KID_BIPED_L_FOOT:
      case KID_BIPED_L_TOE0:
      case KID_BIPED_R_THIGH:
      case KID_BIPED_R_CALF:
      case KID_BIPED_R_FOOT:
      case KID_BIPED_R_TOE0:
        new = reassign_bone_id_kid_to_tac(old);
        if (new < 0 || new == BONE_FIRST_LINK) {
          ERROR("bogus bone reassignment from %u to %d\n", old, new);
          return i + 1;
        }
        gsm->assignments[i]->bone_id = new;
        break;
      default:
        ERROR("WARNING: unexpected bone_id %u in assignment %u\n", gsm->assignments[i]->bone_id, i);
        return i + 1;
    }
    TRACE1("assignment %u converted from %02u %s to %02u %s\n", i, old, get_bone_by_id(gsm, old), gsm->assignments[i]->bone_id, get_biped_bone_name(GSM_TYPE_TAC, gsm->assignments[i]->bone_id));
    gripper_id = get_gripper_by_bone(gsm, old);
    if (gripper_id > -1) gsm->grippers[gripper_id]->bone_id = gsm->assignments[i]->bone_id;
  }

  return 0;
}

/* renumber bones in assignments */
int reassign_swt_to_bf1(struct gsm *gsm) {
  unsigned short i, j, old;
  int new, gripper_id;

  if (gsm->num_assignments == 1) {
    TRACE("converting one assignment\n");
  }
  else TRACE("converting %u assignments\n", gsm->num_assignments);

  for (i = 0; i < gsm->num_assignments; i++) {
    old = gsm->assignments[i]->bone_id;
    switch (old) {
      /* extra fingers unique to SWT */
      case SWT_BIPED_L_FINGER02:
      case SWT_BIPED_L_FINGER12:
      case SWT_BIPED_L_FINGER22:
      case SWT_BIPED_L_FINGER3:
      case SWT_BIPED_L_FINGER31:
      case SWT_BIPED_L_FINGER32:
      case SWT_BIPED_L_FINGER4:
      case SWT_BIPED_L_FINGER41:
      case SWT_BIPED_L_FINGER42:
      case SWT_BIPED_R_FINGER02:
      case SWT_BIPED_R_FINGER12:
      case SWT_BIPED_R_FINGER22:
      case SWT_BIPED_R_FINGER3:
      case SWT_BIPED_R_FINGER31:
      case SWT_BIPED_R_FINGER32:
      case SWT_BIPED_R_FINGER4:
      case SWT_BIPED_R_FINGER41:
      case SWT_BIPED_R_FINGER42:
        /* link vertices to last valid bone */
        if (gsm->assignments[i]->num_vertices) {
          gsm->assignments[i - 1]->num_vertices += gsm->assignments[i]->num_vertices;
          TRACE("old bone %u was %s: vertices moved to %s: %u\n", i, get_biped_bone_name(GSM_TYPE_SWT, old), get_biped_bone_name(GSM_TYPE_BF1, gsm->assignments[i - 1]->bone_id), gsm->assignments[i]->num_vertices);
        }
        /* free this unnecessary assignment */
        TRACE("freeing unneeded assignment %u\n", i);
        free(gsm->assignments[i]);
        gsm->num_assignments--;
        for (j = i; j < gsm->num_assignments; j++) {
          gsm->assignments[j] = gsm->assignments[j + 1];
        }
        /* fall through to next case to deal with this one */
        i--;
        continue;
      case SWT_BIPED_PELVIS:
      case SWT_BIPED_SPINE:
      case SWT_BIPED_SPINE1:
      case SWT_BIPED_SPINE2:
      case SWT_BIPED_NECK:
      case SWT_BIPED_HEAD:
      case SWT_BIPED_L_CLAVICLE:
      case SWT_BIPED_L_UPPERARM:
      case SWT_BIPED_L_FOREARM:
      case SWT_BIPED_L_HAND:
      case SWT_BIPED_L_FINGER0:
      case SWT_BIPED_L_FINGER01:
      case SWT_BIPED_L_FINGER1:
      case SWT_BIPED_L_FINGER11:
      case SWT_BIPED_L_FINGER2:
      case SWT_BIPED_L_FINGER21:
      case SWT_BIPED_R_CLAVICLE:
      case SWT_BIPED_R_UPPERARM:
      case SWT_BIPED_R_FOREARM:
      case SWT_BIPED_R_HAND:
      case SWT_BIPED_R_FINGER0:
      case SWT_BIPED_R_FINGER01:
      case SWT_BIPED_R_FINGER1:
      case SWT_BIPED_R_FINGER11:
      case SWT_BIPED_R_FINGER2:
      case SWT_BIPED_R_FINGER21:
      case SWT_BIPED_L_THIGH:
      case SWT_BIPED_L_CALF:
      case SWT_BIPED_L_FOOT:
      case SWT_BIPED_L_TOE0:
      case SWT_BIPED_R_THIGH:
      case SWT_BIPED_R_CALF:
      case SWT_BIPED_R_FOOT:
      case SWT_BIPED_R_TOE0:
        new = reassign_bone_id_swt_to_bf1(old);
        if (new < 0 || new == BONE_FIRST_LINK) {
          ERROR("bogus bone reassignment from %u to %d\n", old, new);
          return i + 1;
        }
        gsm->assignments[i]->bone_id = new;
        break;
      default:
        ERROR("WARNING: unexpected bone_id %u in assignment %u\n", gsm->assignments[i]->bone_id, i);
        return i + 1;
    }
    TRACE1("assignment %u converted from %02u %s to %02u %s\n", i, old, get_bone_by_id(gsm, old), gsm->assignments[i]->bone_id, get_biped_bone_name(GSM_TYPE_BF1, gsm->assignments[i]->bone_id));
    gripper_id = get_gripper_by_bone(gsm, old);
    if (gripper_id > -1) gsm->grippers[gripper_id]->bone_id = gsm->assignments[i]->bone_id;
  }

  return 0;
}

/* renumber bones in assignments */
int reassign_kid_to_bf1(struct gsm *gsm) {
  unsigned short i, j, old;
  int new, gripper_id;

  if (gsm->num_assignments == 1) {
    TRACE("converting one assignment\n");
  }
  else TRACE("converting %u assignments\n", gsm->num_assignments);

  for (i = 0; i < gsm->num_assignments; i++) {
    old = gsm->assignments[i]->bone_id;
    switch (old) {
      /* kids have 6 ponytails */
      case KID_BIPED_PONYTAIL1:
      case KID_BIPED_PONYTAIL11:
      case KID_BIPED_PONYTAIL12:
      case KID_BIPED_PONYTAIL2:
      case KID_BIPED_PONYTAIL21:
      case KID_BIPED_PONYTAIL22:
        /* link vertices to last valid bone */
        if (gsm->assignments[i]->num_vertices) {
          gsm->assignments[i - 1]->num_vertices += gsm->assignments[i]->num_vertices;
          TRACE("old bone %u was %s: vertices moved to %s: %u\n", i, get_biped_bone_name(GSM_TYPE_KID, old), get_biped_bone_name(GSM_TYPE_BF1, gsm->assignments[i - 1]->bone_id), gsm->assignments[i]->num_vertices);
        }
        /* free this unnecessary assignment */
        TRACE("freeing unneeded assignment %u\n", i);
        free(gsm->assignments[i]);
        gsm->num_assignments--;
        for (j = i; j < gsm->num_assignments; j++) {
          gsm->assignments[j] = gsm->assignments[j + 1];
        }
        /* fall through to next case to deal with this one */
        i--;
        continue;
      case KID_BIPED_PELVIS:
      case KID_BIPED_SPINE:
      case KID_BIPED_SPINE1:
      case KID_BIPED_SPINE2:
      case KID_BIPED_NECK:
      case KID_BIPED_HEAD:
      case KID_BIPED_L_CLAVICLE:
      case KID_BIPED_L_UPPERARM:
      case KID_BIPED_L_FOREARM:
      case KID_BIPED_L_HAND:
      case KID_BIPED_L_FINGER0:
      case KID_BIPED_L_FINGER01:
      case KID_BIPED_L_FINGER1:
      case KID_BIPED_L_FINGER11:
      case KID_BIPED_L_FINGER2:
      case KID_BIPED_L_FINGER21:
      case KID_BIPED_R_CLAVICLE:
      case KID_BIPED_R_UPPERARM:
      case KID_BIPED_R_FOREARM:
      case KID_BIPED_R_HAND:
      case KID_BIPED_R_FINGER0:
      case KID_BIPED_R_FINGER01:
      case KID_BIPED_R_FINGER1:
      case KID_BIPED_R_FINGER11:
      case KID_BIPED_R_FINGER2:
      case KID_BIPED_R_FINGER21:
      case KID_BIPED_L_THIGH:
      case KID_BIPED_L_CALF:
      case KID_BIPED_L_FOOT:
      case KID_BIPED_L_TOE0:
      case KID_BIPED_R_THIGH:
      case KID_BIPED_R_CALF:
      case KID_BIPED_R_FOOT:
      case KID_BIPED_R_TOE0:
        new = reassign_bone_id_kid_to_bf1(old);
        if (new < 0 || new == BONE_FIRST_LINK) {
          ERROR("bogus bone reassignment from %u to %d\n", old, new);
          return i + 1;
        }
        gsm->assignments[i]->bone_id = new;
        break;
      default:
        ERROR("WARNING: unexpected bone_id %u in assignment %u\n", gsm->assignments[i]->bone_id, i);
        return i + 1;
    }
    TRACE1("assignment %u converted from %02u %s to %02u %s\n", i, old, get_bone_by_id(gsm, old), gsm->assignments[i]->bone_id, get_biped_bone_name(GSM_TYPE_BF1, gsm->assignments[i]->bone_id));
    gripper_id = get_gripper_by_bone(gsm, old);
    if (gripper_id > -1) gsm->grippers[gripper_id]->bone_id = gsm->assignments[i]->bone_id;
  }

  return 0;
}

/* renumber bones in assignments */
int reassign_tac_to_bf1(struct gsm *gsm) {
  unsigned short i, j, old;
  int new, gripper_id;

  if (gsm->num_assignments == 1) {
    TRACE("converting one assignment\n");
  }
  else TRACE("converting %u assignments\n", gsm->num_assignments);

  for (i = 0; i < gsm->num_assignments; i++) {
    old = gsm->assignments[i]->bone_id;
    switch (old) {
      case TAC_BIPED_PONYTAIL1:
      case TAC_BIPED_PONYTAIL11:
      case TAC_BIPED_PONYTAIL12:
        /* link vertices to last valid bone */
        if (gsm->assignments[i]->num_vertices) {
          gsm->assignments[i - 1]->num_vertices += gsm->assignments[i]->num_vertices;
          TRACE("old bone %u was %s: vertices moved to %s: %u\n", i, get_biped_bone_name(GSM_TYPE_TAC, old), get_biped_bone_name(GSM_TYPE_BF1, gsm->assignments[i - 1]->bone_id), gsm->assignments[i]->num_vertices);
        }
        /* free this unnecessary assignment */
        TRACE("freeing unneeded assignment %u\n", i);
        free(gsm->assignments[i]);
        gsm->num_assignments--;
        for (j = i; j < gsm->num_assignments; j++) {
          gsm->assignments[j] = gsm->assignments[j + 1];
        }
        /* fall through to next case to deal with this one */
        i--;
        continue;
      case TAC_BIPED_PELVIS:
      case TAC_BIPED_SPINE:
      case TAC_BIPED_SPINE1:
      case TAC_BIPED_SPINE2:
      case TAC_BIPED_NECK:
      case TAC_BIPED_HEAD:
      case TAC_BIPED_L_CLAVICLE:
      case TAC_BIPED_L_UPPERARM:
      case TAC_BIPED_L_FOREARM:
      case TAC_BIPED_L_HAND:
      case TAC_BIPED_L_FINGER0:
      case TAC_BIPED_L_FINGER01:
      case TAC_BIPED_L_FINGER1:
      case TAC_BIPED_L_FINGER11:
      case TAC_BIPED_L_FINGER2:
      case TAC_BIPED_L_FINGER21:
      case TAC_BIPED_R_CLAVICLE:
      case TAC_BIPED_R_UPPERARM:
      case TAC_BIPED_R_FOREARM:
      case TAC_BIPED_R_HAND:
      case TAC_BIPED_R_FINGER0:
      case TAC_BIPED_R_FINGER01:
      case TAC_BIPED_R_FINGER1:
      case TAC_BIPED_R_FINGER11:
      case TAC_BIPED_R_FINGER2:
      case TAC_BIPED_R_FINGER21:
      case TAC_BIPED_L_THIGH:
      case TAC_BIPED_L_CALF:
      case TAC_BIPED_L_FOOT:
      case TAC_BIPED_L_TOE0:
      case TAC_BIPED_R_THIGH:
      case TAC_BIPED_R_CALF:
      case TAC_BIPED_R_FOOT:
      case TAC_BIPED_R_TOE0:
        new = reassign_bone_id_tac_to_bf1(old);
        if (new < 0 || new == BONE_FIRST_LINK) {
          ERROR("bogus bone reassignment from %u to %d\n", old, new);
          return i + 1;
        }
        gsm->assignments[i]->bone_id = new;
        break;
      default:
        ERROR("WARNING: unexpected bone_id %u in assignment %u\n", gsm->assignments[i]->bone_id, i);
        return i + 1;
    }
    TRACE1("assignment %u converted from %02d %s to %02d %s\n", i, old, get_bone_by_id(gsm, old), gsm->assignments[i]->bone_id, get_biped_bone_name(GSM_TYPE_BF1, gsm->assignments[i]->bone_id));
    gripper_id = get_gripper_by_bone(gsm, old);
    if (gripper_id > -1) gsm->grippers[gripper_id]->bone_id = gsm->assignments[i]->bone_id;
  }

  return 0;
}


/* return first_vertex and num_vertices packed up */
unsigned int get_assignments_by_bone(struct gsm *gsm, unsigned short bone_id) {
  unsigned short i;
  unsigned int ret;

  if (! gsm) {
    ERROR("no gsm\n");
    return 0;
  }

  if (bone_id > gsm->skeletons[0]->num_bones) {
    ERROR("bone_id out of range\n");
    return 0;
  }

  ret = 0;
  for (i = 0; i < gsm->num_assignments; i++) {
    if (gsm->assignments[i]->bone_id == bone_id) {
      ret = gsm->assignments[i]->first_vertex + gsm->assignments[i]->num_vertices * 256;
      break;
    }
  }

  return ret;
}

int get_bone_by_position(struct gsm *gsm, unsigned short position_id) {
  unsigned short i;

  if (! gsm || gsm->num_positions <= position_id) {
    ERROR("no position in gsm!\n");
    return -1;
  }

  for (i = 0; i < gsm->num_assignments; i++) {
    if (position_id >= gsm->assignments[i]->first_vertex && position_id <= gsm->assignments[i]->first_vertex + gsm->assignments[i]->num_vertices) return gsm->assignments[i]->bone_id;
  }

  return -1;
} 

/* assignments - don't normally list them as bones took care of it */
int describe_assignments(struct gsm *gsm) {
  unsigned short i;
  struct assignment *assignment;

  if (! gsm) {
    ERROR("no GSM!\n");
    return 1;
  }

  SAY("Vertex to bone assignments: %u\n", gsm->num_assignments);
  if (gsm_verbose > 2) {
    for (i = 0; i < gsm->num_assignments; i++) {
      assignment = gsm->assignments[i];
      if (assignment->num_vertices == 0) {
        SAY("  Assignment %u: vertex %u with bone %u \"%s\"\n", i, assignment->first_vertex, assignment->bone_id, get_bone_by_id(gsm, assignment->bone_id));
      }
      else {
        SAY("  Assignment %u: %u vertices from %u to %u with bone %u \"%s\"\n", i, assignment->num_vertices, assignment->first_vertex, assignment->first_vertex + assignment->num_vertices - 1, assignment->bone_id, get_bone_by_id(gsm, assignment->bone_id));
      }
    }
  }

  return 0;
}
