#include "gsmhacking.h"

struct map **alloc_maps(unsigned short n) {
  return (struct map **) alloc_lots(n, "map", sizeof(struct map *), (void *(*)()) alloc_map);
}

struct map *alloc_map() {
  return (struct map *) alloc_one("map", sizeof(struct map));
}

void cleanup_maps(unsigned short num_maps, struct map **maps) {
  free_structs("map", num_maps, (void **) maps);
}

int find_maps(struct gsm *gsm, struct material *material) {
  unsigned short i;
  struct map *map;

  if (! material->num_maps) return 0;

  if (! gsm || ! gsm->fd) {
    ERROR("no file descriptor for gsm!\n");
    return 1;
  }

  /* allocate a big enough struct */
  material->maps = alloc_maps(material->num_maps);
  if (! material->maps) return 1;

  /* fill in the maps */
  for (i = 0; i < material->num_maps; i++) {
    map = material->maps[i];
    if (parse_map(gsm, i, map, material->num_maps)) {
      ERROR("expected map id %u but got null\n", i);
      return 2;
    }
  }

  return 0;
}

int parse_map(struct gsm *gsm, unsigned short num, struct map *map, unsigned short total) {
  /* find map type */
  if (read_short(gsm->fd, &map->map_type)) {
    ERROR("can't read map type");
    return 1;
  }

  /* read texture number */
  if (read_short(gsm->fd, &map->texture_id)) {
    ERROR("can't read map's texture id");
    return 2;
  }

  TRACE("map %u is %s map of %s\n", num, get_map_type_name(map->map_type), gsm->textures[map->texture_id]->basename);

  return 0;
}

char *get_map_type_name(unsigned short map_type) {
  /* sanity check for map */
  switch (map_type) {
    case MAP_TYPE_AMBIENT:
    case MAP_TYPE_DIFFUSE:
    case MAP_TYPE_SPECULAR:
    case MAP_TYPE_SHININESS:
    case MAP_TYPE_SHINE_STRENGTH:
    case MAP_TYPE_SELF_ILLUMINATION:
    case MAP_TYPE_TRANSPARENCY:
    case MAP_TYPE_FILTER_COLOUR:
    case MAP_TYPE_BUMP:
    case MAP_TYPE_REFLECTION:
    case MAP_TYPE_REFRACTION:
    case MAP_TYPE_DISPLACEMENT:
      return map_type_name[map_type];
  }
  WARN("unknown material type 0x%02x\n", map_type);
  return "unknown type";
}

int get_map_type_by_name(char *name) {
  unsigned short i;

  for (i = 0; i < NUM_MAP_TYPES; i++) {
    if (str_equiv(name, map_type_name[i])) return i;
  }
  /* :biggrin: */
  if (str_equiv(name, "filter-color")) return MAP_TYPE_FILTER_COLOUR;
  WARN("unknown material type name %s\n", name);

  return -1;
}

char *get_all_map_type_names(struct material *material, int *need_free) {
  unsigned short i;
  unsigned int len;
  char *name, *ret, *s;

  *need_free = 0;

  if (! material) {
    ERROR("no material\n");
    return "bogus";
  }

  if (! material->num_maps) return get_map_type_name(material->map_type);

  ret = (char *) malloc(sizeof(char) * 15 * material->num_maps);
  if (! ret) {
    ERROR("out of memory for map names\n");
    return "unknown";
  }
  *need_free = 1;

  s = ret;
  TRACE("getting names: %u\n", material->num_maps);
  for (i = 0; i < material->num_maps; i++) {
    name = get_map_type_name(material->maps[i]->map_type);
    TRACE1("writing name %u %s to %p\n", i, name, s);
    len = snprintf(s, 15, "%s", name);
    TRACE2("snprintf({s=%p}, 15, \"%s\"): %u\n", s, name, len);
    if (len == -1) {
      ERROR("snprintf() failed for map name %u\n", i);
      free(ret);
      *need_free = 0;
      return "unknown";
    }
    s += len;
    if (i < material->num_maps - 1) {
      if (i < material->num_maps - 2) {
        TRACE1("writing comma %u to %p\n", i, s);
        len = snprintf(s, 3, ", ");
        TRACE2("snprintf({s=%p}, 3, \", \"): %u\n", s, len);
        if (len == -1) {
          ERROR("snprintf() failed for comma %u\n", i);
          free(ret);
          *need_free = 0;
          return "unknown";
        }
      }
      else {
        TRACE1("writing \"and\" %u to %p\n", i, s);
        len = snprintf(s, 6, " and ");
        TRACE2("snprintf({s=%p}, 6, \" and \"): %u\n", s, len);
        if (len == -1) {
          ERROR("snprintf() failed for \"and\" %u\n", i);
          free(ret);
          *need_free = 0;
          return "unknown";
        }
      }
    }
    TRACE2("string ret at %p is \"%s\"\n", ret, ret);
    s += len;
  }

  return ret;
}

char *get_all_map_type_names_with_textures(struct material *material, int *need_free) {
  unsigned short i;
  unsigned int len;
  char *name, *map, *ret, *s;

  *need_free = 0;

  if (! material->num_maps) {
    ret = (char *) malloc(sizeof(char) * 29);
    if (! ret) {
      ERROR("out of memory for map names\n");
      return "unknown";
    }
    map = get_map_type_name(material->map_type);
    if (snprintf(ret, 29, "%s (no textures)", map) == -1) {
      ERROR("snprintf() barfed");
      free(ret);
      *need_free = 0;
      return "broken";
    }
    *need_free = 1;
    return ret;
  }

  ret = (char *) malloc(sizeof(char) * 23 * material->num_maps);
  if (! ret) {
    ERROR("out of memory for map names\n");
    return "unknown";
  }

  s = ret;
  TRACE("getting names: %u\n", material->num_maps);
  for (i = 0; i < material->num_maps; i++) {
    name = get_map_type_name(material->maps[i]->map_type);
    map = get_texture_id(material->maps[i], need_free);
    TRACE1("writing name %u %s to %p\n", i, name, s);
    len = snprintf(s, 23, "%s (%s)", name, map);
    if (*need_free) free(map);
    *need_free = 0;
    if (len == -1) {
      ERROR("snprintf() failed for map name %u\n", i);
      free(ret);
      return "unknown";
    }
    s += len;
    if (i < material->num_maps - 1) {
      if (i < material->num_maps - 2) {
        TRACE1("writing comma %u to %p\n", i, s);
        len = snprintf(s, 3, ", ");
        TRACE2("snprintf({s=%p}, 3, \", \"): %u\n", s, len);
        if (len == -1) {
          ERROR("snprintf() failed for comma %u\n", i);
          free(ret);
          return "unknown";
        }
      }
      else {
        TRACE1("writing \"and\" %u to %p\n", i, s);
        len = snprintf(s, 6, " and ");
        TRACE2("snprintf({s=%p}, 6, \" and \"): %u\n", s, len);
        if (len == -1) {
          ERROR("snprintf() failed for \"and\" %u\n", i);
          free(ret);
          return "unknown";
        }
      }
    }
    *need_free = 1;
    TRACE2("string ret at %p is \"%s\"\n", ret, ret);
    s += len;
  }

  return ret;
}

int write_maps(struct gsm *gsm, unsigned short num) {
  unsigned short i;

  if (! gsm || ! gsm->out_fd) {
    ERROR("no output file descriptor for gsm!\n");
    return 1;
  }

  TRACE("writing number of maps: %u\n", gsm->materials[num]->num_maps);
  if (write_short(gsm->out_fd, &gsm->materials[num]->num_maps)) {
    ERROR("can't write material's number of maps");
    return 1;
  }

  /* nothing more to be done */
  if (! gsm->materials[num]->num_maps) return 0;

  DEBUG("writing maps for material %u: %u\n", num, gsm->materials[num]->num_maps);
  for (i = 0; i < gsm->materials[num]->num_maps; i++) {
    /* write map type */
    if (write_short(gsm->out_fd, &gsm->materials[num]->maps[i]->map_type)) {
      ERROR("can't write map %u type", i);
      return 1;
    }
  
    /* write texture number */
    if (write_short(gsm->out_fd, &gsm->materials[num]->maps[i]->texture_id)) {
      ERROR("can't write map %u's texture id", i);
      return 2;
    }
  }

  return 0;
}

unsigned short get_num_maps(struct gsm *gsm, unsigned short texture) {
  unsigned short i, j, count;

  if (! gsm || gsm->num_textures < texture) {
    ERROR("no texture in gsm!\n");
    return -1;
  }

  count = 0;
  for (i = 0; i < gsm->num_materials; i++) {
    for (j = 0; j < gsm->materials[i]->num_maps; j++) {
      if (gsm->materials[i]->maps[j]->texture_id == texture) count++;
    }
  }

  return count;
}

struct map_set *alloc_map_set() {
  return (struct map_set *) alloc_one("map_set", sizeof(struct map_set));
}

int add_to_map_set(struct map_set **map_sets, char *material, char *map_type, char *texture_arg) {
  unsigned short len;
  struct map_set *last, *tmp;
  char *bogus;

  tmp = alloc_map_set();
  if (! tmp) {
    ERROR("out of memory for map_set");
    return 1;
  }
  TRACE("tmp: %p\n", tmp);
  tmp->material = alloc_name(material, &len);
  if (! len) {
    ERROR("out of memory for map_set material");
    return 2;
  }

  if (map_type) {
    tmp->type = alloc_name(map_type, &len);
    if (! len) {
      ERROR("out of memory for map_set map type");
      return 3;
    }
  }
  else tmp->type = 0;

  if (texture_arg) {
    tmp->texture_id = (unsigned short) strtol(texture_arg, &bogus, 0);
    if (*bogus) {
      ERROR("bogus texture ID in map set\n");
      return 4;
    }
  }
  else tmp->texture_id = -1;
  TRACE("tmp->material: %p, tmp->type: %p, tmp->texture_id: %u\n", tmp->material, tmp->type, tmp->texture_id);

  last = (struct map_set *) gll_last((struct gll *) *map_sets);
  TRACE("last: %p\n", last);
  if (last) last->next = tmp;
  else *map_sets = tmp;
  /*gll_debug((struct gll *) *map_sets);*/

  return 0;
}

int set_maps(struct gsm *gsm) {
  struct map_set *map_set;
  struct material *material;
  struct map *map, **new_maps;
  unsigned short expected, replaced;
  unsigned short i, material_id, map_type, num_maps;
  int len, texture_id;
  char *bogus;

  if (! gsm || ! gsm->num_materials) {
    ERROR("no materials in gsm\n");
    return -1;
  }

  if (! gsm->option->map_sets) {
    ERROR("no set parameters\n");
    return -2;
  }

  expected = replaced = 0;
  GLL_LOOP(map_set, gsm->option->map_sets) {
    expected++;

    /* first parse the options */
    /* material id can be a number (preferred) or a material name */
    material_id = (unsigned short) strtol(map_set->material, &bogus, 0);
    if (*bogus) {
      len = get_material_by_name(gsm, map_set->material);
      if (len < 0) {
        error_dbg("bogus material in map set\n");
        continue;
      }
      material_id = len;
    }
    if (material_id > gsm->num_materials) {
      error_dbg("material id %u out of range\n", material_id);
      continue;
    }
    material = gsm->materials[material_id];

    /* check valid texture */
    texture_id = map_set->texture_id;
    if (texture_id > gsm->num_textures) {
      error_dbg("texture id %u out of range\n", texture_id);
      continue;
    }

    /* texture -1 => --clear-map was called */
    if (texture_id == -1) {
      replaced++;
      SAY("clearing maps from material %02u: %s\n", material_id, material->name);
      cleanup_maps(material->num_maps, material->maps);
      material->num_maps = 0;
      continue;
    }

    /* map type can be a number or name */
    map_type = (unsigned short) strtol(map_set->type, &bogus, 0);
    if (*bogus) {
      len = get_map_type_by_name(map_set->type);
      if (len < 0) {
        error_dbg("bogus map type in map set\n");
        continue;
      }
      map_type = len;
    }
    if (map_type > NUM_MAP_TYPES) {
      error_dbg("map type %u out of range\n", map_type);
      continue;
    }

    DEBUG("making %s map with texture %s for material %s\n", get_map_type_name(map_type), gsm->textures[texture_id]->basename, material->name);

    /* make the new map */
    map = alloc_map();
    if (! map) {
      ERROR("out of memory for new map");
      return -3;
    }
    map->map_type = map_type;
    map->texture_id = texture_id;

    /* if we have already been called, append us to the map list */
    if (material->changed) {
      num_maps = material->num_maps + 1;
      new_maps = alloc_maps(num_maps);
      for (i = 0; i < material->num_maps; i++) {
        memmove(new_maps[i], material->maps[i], sizeof(struct map));
      }
    }
    else {
      num_maps = 1;
      new_maps = alloc_maps(num_maps);
      material->num_maps = 0;
    }
    new_maps[material->num_maps] = map;
    cleanup_maps(material->num_maps, material->maps);
    material->maps = new_maps;
    material->num_maps = num_maps;
    material->changed = 1;

    /* cool, it worked */
    replaced++;
  }


  if (replaced == 1) SAY("set one map\n");
  else if (replaced) SAY("set %u maps\n", replaced);
  else if (replaced != expected) error_dbg("no maps set, expected to set %u\n", expected);
  else error_dbg("no maps set\n");

  return replaced;
}

int list_map_types() {
  unsigned short i;

  SAY("map types understood by gsmhacking (for use with --map-texture etc)\n");
  for (i = 0; i < NUM_MAP_TYPES; i++) SAY("%2u: %s\n", i, get_map_type_name(i));

  return 1;
}
