#ifndef MATERIAL_H
#define MATERIAL_H
#include <stdlib.h>
#include "data.h"

#define MATERIAL_PROPERTY_AMBIENT MAP_TYPE_AMBIENT
#define MATERIAL_PROPERTY_DIFFUSE MAP_TYPE_DIFFUSE
#define MATERIAL_PROPERTY_SPECULAR MAP_TYPE_SPECULAR
#define MATERIAL_PROPERTY_SHININESS MAP_TYPE_SHININESS
#define MATERIAL_PROPERTY_SHINE_STRENGTH MAP_TYPE_SHINE_STRENGTH
#define MATERIAL_PROPERTY_SELF_ILLUMINATION MAP_TYPE_SELF_ILLUMINATION
#define MATERIAL_PROPERTY_TRANSPARENCY MAP_TYPE_TRANSPARENCY
#define MATERIAL_PROPERTY_FILTER_COLOUR MAP_TYPE_FILTER_COLOUR
#define MATERIAL_PROPERTY_BUMP MAP_TYPE_BUMP
#define MATERIAL_PROPERTY_REFLECTION MAP_TYPE_REFLECTION
#define MATERIAL_PROPERTY_REFRACTION MAP_TYPE_REFRACTION
#define MATERIAL_PROPERTY_DISPLACEMENT MAP_TYPE_DISPLACEMENT
#define MATERIAL_PROPERTY_NAME NUM_MAP_TYPES + 0

struct material_set {
  struct material_set *next;
  char *material;
  unsigned short property; /* MAP_TYPE_TRANSPARENCY means edit transparency */
  int value1; /* R for RGB stuff, the number for transparency etc */
  int value2; /* G for RGB stuff */
  int value3; /* B for RGB stuff */
  char *name;
};

GSM_API struct material **alloc_materials(unsigned short);
GSM_API struct material **alloc_more_materials(unsigned short, unsigned short, struct material **);
GSM_API struct material *alloc_material();
GSM_API void cleanup_materials(unsigned short, struct material **);
GSM_API int find_materials(struct gsm *);
GSM_API int parse_material(struct gsm *, unsigned short, struct material *, unsigned short);
GSM_API int write_materials(struct gsm *);
GSM_API int get_material_by_name(struct gsm *, char *);
GSM_API char *get_material_name(struct gsm *, unsigned short);
GSM_API char *get_texture_id(struct map *, int *);
GSM_API char *get_texture_ids(struct material *, int *);
GSM_API struct material_set *alloc_material_set();
GSM_API int add_to_material_set(struct material_set **, unsigned short, char *, char *, char *, char *);
GSM_API int set_material_properties(struct gsm *);
GSM_API int describe_materials(struct gsm *);

struct material {
  unsigned short number;
  unsigned int offset;
  unsigned short name_length;
  char *name;
  unsigned short num_maps;
  unsigned short map_type;
  float diffuse_red;
  float diffuse_green;
  float diffuse_blue;
  float ambient_red;
  float ambient_green;
  float ambient_blue;
  float specular_red;
  float specular_green;
  float specular_blue;
  float opacity_falloff;
  float transparency;
  char twosided;
  float self_illumination;
  float shininess;
  float shine_strength;
  struct map **maps;
  unsigned char changed; /* set when we use a map_set */
};

#endif
