#include "gsmhacking.h"
#include <math.h>

/* large parts of code copied almost verbatim from XXX where? */

vector_t **alloc_vectors(unsigned short n) {
  return (vector_t **) alloc_lots(n, "vector", sizeof(vector_t *), (void *(*)()) alloc_vector);
}

vector_t *alloc_vector() {
  return (vector_t *) alloc_one("vector", sizeof(vector_t));
}

void cleanup_vectors(unsigned short num_vectors, vector_t **vectors) {
 free_structs("vector", num_vectors, (void *) vectors);
}

matrix_t **alloc_matrices(unsigned short n) {
  return (matrix_t **) alloc_lots(n, "matrix", sizeof(matrix_t *), (void *(*)()) alloc_matrix);
}

matrix_t *alloc_matrix() {
  matrix_t *ret;
  ret = (matrix_t *) alloc_one("matrix", sizeof(matrix_t));
  if (! ret) return 0;
  (*ret)[R0] = (*ret)[R4] = (*ret)[R8] = 1;
  (*ret)[3] = (*ret)[7] = (*ret)[11] = (*ret)[15] = 1;
  return ret;
}

void cleanup_matrices(unsigned short num_matrices, matrix_t **matrices) {
 free_structs("matrix", num_matrices, (void *) matrices);
}

vector_t *scalar_product(float x, vector_t *v) {
  vector_t *ret;

  if (! v) return 0;

  ret = alloc_vector();
  if (! ret) return 0;

  ret->x = x * v->x;
  ret->y = x * v->y;
  ret->z = x * v->z;

  return ret;
}

float dot_product(vector_t *u, vector_t *v) {
  if (! u) return 0;
  if (! v) return 0;

  return (float) ((double) u->x * (double) v->x + (double) u->y * (double) v->y + (double) u->z * (double) v->z);
}

vector_t *cross_product(vector_t *u, vector_t *v) {
  vector_t *ret;

  if (! u) return 0;
  if (! v) return 0;

  ret = alloc_vector();
  if (! ret) return 0;

  ret->x = (float) ((double) (u->y * v->z) - (double) (u->z * v->y));
  ret->y = (float) ((double) (u->z * v->x) - (double) (u->x * v->z));
  ret->z = (float) ((double) (u->x * v->y) - (double) (u->y * v->x));

  return ret;
}

float vector_magnitude(vector_t *v) {
  if (! v) return -1;
  return (float) sqrt((double) (v->x * v->x) + (double) (v->y * v->y) + (double) (v->z * v->z));
}

int normalise_vector(vector_t *v) {
  vector_t *ret;
  if (! v) return 1;
  ret = scalar_product(1 / vector_magnitude(v), v);
  if (! ret) return 2;
  memmove(v, ret, sizeof(vector_t));
  free(ret);
  return 0;
}

int vector_equiv(vector_t *v, vector_t *u) {
  if (! v || !u) {
    ERROR("where are the vectors?\n");
    return 0;
  }

  if (fabs(v->x - u->x) > FLOAT_EQUIV_THRESHOLD) return 0;
  if (fabs(v->y - u->y) > FLOAT_EQUIV_THRESHOLD) return 0;
  if (fabs(v->z - u->z) > FLOAT_EQUIV_THRESHOLD) return 0;

  return 1;
}

vector_t *get_vector_by_positions(struct position *p, struct position *q) {
  vector_t *ret = alloc_vector();
  if (! ret) return 0;

  ret->x = p->local.x - q->local.x;
  ret->y = p->local.y - q->local.y;
  ret->z = p->local.z - q->local.z;

  return ret;
}

matrix_t *make_rotation_matrix_in_x(float radians) {
  matrix_t *ret;
  double c, s;

  ret = alloc_matrix();
  if (! ret) return 0;

  c = cos((double) radians);
  s = sin((double) radians);

  (*ret)[R0] = 1;
  (*ret)[R1] = 0;
  (*ret)[R2] = 0;

  (*ret)[R3] = 0;
  (*ret)[R4] = (float) c;
  (*ret)[R5] = (float) s;

  (*ret)[R6] = 0;
  (*ret)[R7] = (float) -s;
  (*ret)[R8] = (float) c;

  return ret;
}

matrix_t *make_rotation_matrix_in_y(float radians) {
  matrix_t *ret;
  double c, s;

  ret = alloc_matrix();
  if (! ret) return 0;

  c = cos((double) radians);
  s = sin((double) radians);

  (*ret)[R0] = (float) c;
  (*ret)[R1] = 0;
  (*ret)[R2] = (float) -s;

  (*ret)[R3] = 0;
  (*ret)[R4] = 1;
  (*ret)[R5] = 0;

  (*ret)[R6] = (float) s;
  (*ret)[R7] = 0;
  (*ret)[R8] = (float) c;

  return ret;
}

matrix_t *make_rotation_matrix_in_z(float radians) {
  matrix_t *ret;
  double c, s;

  ret = alloc_matrix();
  if (! ret) return 0;

  c = cos((double) radians);
  s = sin((double) radians);

  (*ret)[R0] = (float) c;
  (*ret)[R1] = (float) s;
  (*ret)[R2] = 0;

  (*ret)[R3] = (float) -s;
  (*ret)[R4] = (float) c;
  (*ret)[R5] = 0;

  (*ret)[R6] = 0;
  (*ret)[R7] = 0;
  (*ret)[R8] = 1;

  return ret;
}

/* convert a UNIT quaternion to a rotation matrix */
int quaternion_to_matrix(matrix_t *m, quaternion_t *q) {
  if (! m) return 1;

  /* ww + xx + yy + zz = 1 because ||q|| = 1 */
  M[R0] = 1 - (2 * (q->y * q->y + q->z * q->z));
  M[R1] = (float) (double) (2 * ((double) (q->x * q->y) + (double) (q->z * q->w)));
  M[R2] = (float) (double) (2 * ((double) (q->x * q->z) - (double) (q->y * q->w)));
  M[3] = 0;

  M[R3] = (float) (double) (2 * ((double) (q->x * q->y) - (double) (q->z * q->w)));
  M[R4] = 1 - (2 * (q->x * q->x + q->z * q->z));
  M[R5] = (float) (double) (2 * ((double) (q->y * q->z) + (double) (q->x * q->w)));
  M[7] = 0;

  M[R6] = (float) (double) (2 * ((double) (q->x * q->z) + (double) (q->y * q->w)));
  M[R7] = (float) (double) (2 * ((double) (q->y * q->z) - (double) (q->x * q->w)));
  M[R8] = 1 - (2 * (q->x * q->x + q->y * q->y));
  M[11] = 0;

  M[TX] = 0;
  M[TY] = 0;
  M[TZ] = 0;
  M[15] = 1;

  debug_matrix(m);

  return 0;
}

#define swap_elements(x, y) tmp = M[x]; M[x] = M[y]; M[y] = tmp
int transpose_matrix(matrix_t *m) {
  float tmp;

  if (! m) return 1;

  swap_elements(1, 4);
  swap_elements(2, 8);
  swap_elements(3, 12);

  swap_elements(6, 9);
  swap_elements(7, 13);

  swap_elements(11, 14);

  debug_matrix(m);

  return 0;
}

void debug_matrix(matrix_t *m) {
  if (! m) return;
  DEBUG("| % 0.4f, % 0.4f, % 0.4f, % 0.4f  |\n", M[R0], M[R3], M[R6], M[TX]);
  DEBUG("| % 0.4f, % 0.4f, % 0.4f, % 0.4f  |\n", M[R1], M[R4], M[R7], M[TY]);
  DEBUG("| % 0.4f, % 0.4f, % 0.4f, % 0.4f  |\n", M[R2], M[R5], M[R8], M[TZ]);
  DEBUG("| % 0.4f, % 0.4f, % 0.4f, % 0.4f  |\n", M[3], M[7], M[11], M[15]);
}

/* right for a 4x4 matrix? */
int matrix_to_euler(euler_t *e, matrix_t *m) {
  float tmp;

  if (! e) return 1;
  if (! m) return 2;

  tmp = fabs(M[2]);
  if (tmp > 0.9999999) {
    (*e)[0] = 0;
    (*e)[1] = (float) (double) (-(PI/2) * M[R2] / tmp);
    (*e)[2] = (float) (double) atan2(-M[R3], -M[R2] * M[R6]);
  }
  else {
    (*e)[0] = (float) (double) atan2(M[R5], M[R8]);
    (*e)[1] = (float) (double) asin(-M[R2]);
    (*e)[2] = (float) (double) atan2(M[R1], M[R0]);
  }

  debug_euler(e);

  return 0;
}

void debug_euler(euler_t *e) {
  if (! e) return;
  DEBUG("( % 0.4f, % 0.4f, % 0.4f )\n", (*e)[0], (*e)[1], (*e)[2]);
}

void debug_vector(vector_t *v) {
  if (! v) return;
  DEBUG("( % 0.4f, % 0.4f, % 0.4f )\n", v->x, v->y, v->z);
}

float quaternion_magnitude(quaternion_t *q) {
  if (! q) return -1;
  return (float) sqrt((double) (q->x * q->x) + (double) (q->y * q->y) + (double) (q->z * q->z) + (double) (q->w * q->w));
}

int normalise_quaternion(quaternion_t *q) {
  float s;
  if (! q) return 1;
  s = quaternion_magnitude(q);
  q->x /= s;
  q->y /= s;
  q->z /= s;
  q->w /= s;
  return 0;
}

quaternion_t *multiply_quaternions(quaternion_t *q, quaternion_t *r) {
  quaternion_t *ret;
  /*
  vector_t qv, rv, *u, *v, *w;
  */

  if (! q) return 0;
  if (! r) return 0;

  ret = alloc_quaternion();
  if (! ret) return 0;

  /* grab vector parts of quaternions */
  /*
  qv.x = q->x; qv.y = q->z; qv.z = q->z;
  rv.x = r->x; rv.y = r->z; rv.z = r->z;

  ret->w = (float) ((double) q->w * (double) r->w - dot_product(&qv, &rv));
  u = scalar_product(q->w, &rv);
  if (! u) return 0;
  v = scalar_product(r->w, &qv);
  if (! v) return 0;
  w = cross_product(&qv, &rv);
  if (! w) return 0;
  ret->x = u->x + v->x - w->x;
  ret->y = u->x + v->x - w->x;
  ret->z = u->x + v->x - w->x;
  */

  ret->w = (float) ((double) q->w * (double) r->w - (double) q->x * (double) r->x - (double) q->y * (double) r->y - q->z * r->z);
  ret->x = (float) ((double) q->w * (double) r->x + (double) q->x * (double) r->w - (double) q->y * (double) r->z - (double) q->z * (double) r->y);
  ret->y = (float) ((double) q->w * (double) r->y + (double) q->y * (double) r->w + (double) q->z * (double) r->x - (double) q->x * (double) r->z);
  ret->z = (float) ((double) q->w * (double) r->z + (double) q->z * (double) r->w + (double) q->x * (double) r->y - (double) q->y * (double) r->x);

  //normalise_quaternion(ret);
  debug_quaternion(ret);

  return ret;
}

quaternion_t *concatenate_quaternions(quaternion_t *q, quaternion_t *r) {
  quaternion_t *ret;
  vector_t u, v, *w;
  float dot;

  if (! q) return 0;
  if (! r) return 0;

  ret = alloc_quaternion();
  if (! ret) return 0;

  u.x = q->x; u.y = q->y; u.z = q->z;
  v.x = r->x; v.y = r->y; v.z = r->z;

  dot = dot_product(&u, &v);
  w = cross_product(&u, &v);
  if (! w) {
    free(ret);
    return 0;
  }

  ret->x = u.x + v.x + w->x;
  ret->y = u.y + v.y + w->y;
  ret->z = u.z + v.z + w->z;
  ret->w = q->w * r->w - dot;
  normalise_quaternion(ret);

  free(w);

  return ret;
}

matrix_t *multiply_matrices(matrix_t *m, matrix_t *n) {
  matrix_t *ret;

  if (! m) return 0;
  if (! n) return 0;

  ret = alloc_matrix();
  if (! ret) return 0;

  (*ret)[R0] = M[R0] * N[R0] + M[R3] * N[R1] + M[R6] * N[R2];
  (*ret)[R1] = M[R1] * N[R0] + M[R4] * N[R1] + M[R7] * N[R2];
  (*ret)[R2] = M[R2] * N[R0] + M[R5] * N[R1] + M[R8] * N[R2];
   (*ret)[3] = 0;

  (*ret)[R3] = M[R0] *  N[R3] + M[R3] * N[R4] + M[R6] * N[R5];
  (*ret)[R4] = M[R1] *  N[R3] + M[R4] * N[R4] + M[R7] * N[R5];
  (*ret)[R5] = M[R2] *  N[R3] + M[R5] * N[R4] + M[R8] * N[R5];
   (*ret)[7] = 0;

  (*ret)[R6] = M[R0] * N[R6] + M[R3] * N[R7] + M[R6] * N[R8];
  (*ret)[R7] = M[R1] * N[R6] + M[R4] * N[R7] + M[R7] * N[R8];
  (*ret)[R8] = M[R2] * N[R6] + M[R5] * N[R7] + M[R8] * N[R8];
  (*ret)[11] = 0;

  (*ret)[TX] = M[R0] * N[TX] + M[R3] * N[TY] + M[R6] * N[TZ];
  (*ret)[TY] = M[R1] * N[TX] + M[R4] * N[TY] + M[R7] * N[TZ];
  (*ret)[TZ] = M[R2] * N[TX] + M[R5] * N[TY] + M[R8] * N[TZ];
  (*ret)[15] = 1;

  /* general case
   (*ret)[0] =  M[0] *  N[0] +  M[4] *  N[1] +  M[8] *  N[2] + M[12] *  N[3];
   (*ret)[1] =  M[1] *  N[0] +  M[5] *  N[1] +  M[9] *  N[2] + M[13] *  N[3];
   (*ret)[2] =  M[2] *  N[0] +  M[6] *  N[1] + M[10] *  N[2] + M[14] *  N[3];
   (*ret)[3] =  M[3] *  N[0] +  M[7] *  N[1] + M[11] *  N[2] + M[15] *  N[3];

   (*ret)[4] =  M[0] *  N[4] +  M[4] *  N[5] +  M[8] *  N[6] + M[12] *  N[7];
   (*ret)[5] =  M[1] *  N[4] +  M[5] *  N[5] +  M[9] *  N[6] + M[13] *  N[7];
   (*ret)[6] =  M[2] *  N[4] +  M[6] *  N[5] + M[10] *  N[6] + M[14] *  N[7];
   (*ret)[7] =  M[4] *  N[4] +  M[7] *  N[5] + M[11] *  N[6] + M[15] *  N[7];

   (*ret)[8] =  M[0] *  N[8] +  M[4] *  N[9] +  M[8] * N[10] + M[12] * N[11];
   (*ret)[9] =  M[1] *  N[8] +  M[5] *  N[9] +  M[9] * N[10] + M[13] * N[11];
  (*ret)[10] =  M[2] *  N[8] +  M[6] *  N[9] + M[10] * N[10] + M[14] * N[11];
  (*ret)[11] =  M[4] *  N[8] +  M[7] *  N[9] + M[11] * N[10] + M[15] * N[11];

  (*ret)[12] =  M[0] * N[12] +  M[4] * N[13] +  M[8] * N[14] + M[12] * N[15];
  (*ret)[13] =  M[1] * N[12] +  M[5] * N[13] +  M[9] * N[14] + M[13] * N[15];
  (*ret)[14] =  M[2] * N[12] +  M[6] * N[13] + M[10] * N[14] + M[14] * N[15];
  (*ret)[15] =  M[4] * N[12] +  M[7] * N[13] + M[11] * N[14] + M[15] * N[15];
  */

  debug_matrix(ret);

  return ret;
}

vector_t *multiply_vector_by_matrix(vector_t *v, matrix_t *m) {
  vector_t *ret;

  if (! v) return 0;
  if (! m) return 0;

  ret = alloc_vector();
  if (! ret) return 0;

  ret->x = (float) (((double) M[R0] * (double) v->x)
                   + ((double) M[R3] * (double) v->y)
                   + ((double) M[R6] * (double) v->z));
  ret->y = (float) (((double) M[R1] * (double) v->x)
                   + ((double) M[R4] * (double) v->y)
                   + ((double) M[R7] * (double) v->z));
  ret->z = (float) (((double) M[R2] * (double) v->x)
                   + ((double) M[R5] * (double) v->y)
                   + ((double) M[R8] * (double) v->z));

  return ret;
}

int quaternion_to_euler(euler_t *e, quaternion_t *q) {
  matrix_t m;

  if (! e) return 1;

  if (normalise_quaternion(q)) return 2;
  if (quaternion_to_matrix(&m, q)) return 3;
  if (matrix_to_euler(e, &m)) return 4;

  return 0;
}

int euler_to_quaternion(quaternion_t *q, euler_t *e) {
  double cy, cp, cr, sy, sp, sr;

  if (! q) return 1;
  if (! e) return 2;

  cy = cos(0.5f * (*e)[2]);
  cp = cos(0.5f * (*e)[1]);
  cr = cos(0.5f * (*e)[0]);
  sy = sin(0.5f * (*e)[2]);
  sp = sin(0.5f * (*e)[1]);
  sr = sin(0.5f * (*e)[0]);

  q->x = (float) (cy * cp * sr - sy * sp * cr);
  q->y = (float) (cy * sp * cr + sy * cp * sr);
  q->z = (float) (sy * cp * cr - cy * sp * sr);
  q->w = (float) (cy * cp * cr + sy * sp * sr);

  normalise_quaternion(q);
  debug_quaternion(q);

  return 0;
}

int euler_to_matrix(matrix_t *m, euler_t *e) {
  quaternion_t q;

  if (! m) return 1;
  if (! e) return 2;

  if (euler_to_quaternion(&q, e)) return 3;
  if (quaternion_to_matrix(m, &q)) return 4;

  return 0;
}

int matrix_to_quaternion(quaternion_t *q, matrix_t *m) {
  //euler_t e;
  float trace;
  double s;

  if (! q) return 1;
  if (! m) return 2;

  /*
  if (matrix_to_euler(&e, m)) return 3;
  if (euler_to_quaternion(q, &e)) return 4;
  */

  trace = M[R0] + M[R4] + M[R8] + 1;

  if (trace > 0) {
    s = 0.5 / sqrt(trace);

    q->w = (float) (0.25 / s);
    q->x = (float) ((M[R5] - M[R7]) * s);
    q->y = (float) ((M[R6] - M[R2]) * s);
    q->z = (float) ((M[R1] - M[R3]) * s);

    return 0;
  }

  if (M[R0] >= M[R4] && M[R0] >= M[R8]) {
    s = 2 * sqrt(1 + M[R0] - M[R4] - M[R8]);

    q->w = (float) ((M[R7] + M[R5]) / s);
    q->x = (float) (0.5 / s);
    q->y = (float) ((M[R3] + M[R1]) / s);
    q->z = (float) ((M[R6] + M[R2]) / s);

    return 0;
  }

  if (M[R4] >= M[R0] && M[R4] >= M[R8]) {
    s = 2 * sqrt(1 + M[R4] - M[R0] - M[R8]);

    q->w = (float) ((M[R6] + M[R2]) / s);
    q->x = (float) ((M[R3] + M[R1]) / s);
    q->y = (float) (0.5 / s);
    q->z = (float) ((M[R7] + M[R5]) / s);

    return 0;
  }

  if (M[R8] >= M[R0] && M[R8] >= M[R0]) {
    s = 2 * sqrt(1 + M[R8] - M[R0] - M[R4]);

    q->w = (float) ((M[R3] + M[R1]) / s);
    q->x = (float) ((M[R6] + M[R2]) / s);
    q->y = (float) ((M[R7] + M[R5]) / s);
    q->z = (float) (0.5 / s);

    return 0;
  }

  ERROR("how did we get here?\n");
  return -1;
}

void debug_quaternion(quaternion_t *q) {
  DEBUG("Q(% 0.4f, (% 0.4f, % 0.4f, % 0.4f)), ||Q||: %f\n", q->w, q->x, q->y, q->z, quaternion_magnitude(q));
}

quaternion_t **alloc_quaternions(unsigned short n) {
  return (quaternion_t **) alloc_lots(n, "quaternion", sizeof(quaternion_t *), (void *(*)()) alloc_quaternion);
}

quaternion_t *alloc_quaternion() {
  return (quaternion_t *) alloc_one("quaternion", sizeof(quaternion_t));
}

void cleanup_quaternions(unsigned short num_quaternions, quaternion_t **quaternions) {
 free_structs("quaternion", num_quaternions, (void *) quaternions);
}

quaternion_t *multiply_quaternion_by_vector(quaternion_t *q, vector_t *v) {
  quaternion_t *ret;

  if (! q) return 0;
  if (! v) return 0;

  ret = alloc_quaternion();
  if (! ret) return 0;

  ret->w = q->x * v->x + q->y * v->y + q->z * v->z;
  ret->x = q->w * v->x + q->y * v->z - q->z * v->y;
  ret->y = q->w * v->y + q->z * v->x - q->x * v->z;
  ret->z = q->w * v->z + q->x * v->y - q->y * v->x;

  return ret;
}

int invert_quaternion(quaternion_t *q) {
  if (! q) return 1;

  q->x = -q->x;
  q->y = -q->y;
  q->z = -q->z;

  return 0;
}

quaternion_t *multiply_vector_by_quaternion(vector_t *v, quaternion_t *q) {
  quaternion_t *ret;

  if (! q) return 0;
  if (! v) return 0;

  ret = alloc_quaternion();
  if (! ret) return 0;

  ret->w = -(q->x * v->x + q->y * v->y + q->z * v->z);
  ret->x = q->w * v->x + q->z * v->y - q->y * v->z;
  ret->y = q->w * v->y + q->x * v->z - q->z * v->x;
  ret->z = q->w * v->z + q->y * v->x - q->x * v->y;

  return ret;
}

vector_t *rotate_vector(vector_t *v, quaternion_t *q) {
  quaternion_t *r, *tmp, c;
  vector_t *ret;

  if (! v) return 0;
  if (! q) return 0;

  tmp = multiply_quaternion_by_vector(q, v);
  if (! tmp) return 0;
  memmove(&c, q, sizeof(quaternion_t));
  if (invert_quaternion(&c)) return 0;
  //normalise_quaternion(tmp);
  r = multiply_quaternions(tmp, &c);
  free(tmp);
  if (! r) return 0;
  //normalise_quaternion(r);

  ret = alloc_vector();
  if (! ret) {
    free(r);
    return 0;
  }

  ret->x = r->x;
  ret->y = r->y;
  ret->z = r->z;

  free(r);

  return ret;
}

float radians_to_degrees(float radians) {
  return (float) (((double) radians * 180) / (PI));
}

float degrees_to_radians(float degrees) {
  return (float) ((double) degrees * ((PI) / 180));
}

/* given rotations src and dest, find q such that src x q = dest */
quaternion_t *find_rotation(quaternion_t *src, quaternion_t *dest) {
  quaternion_t c;

  if (! src) return 0;
  if (! dest) return 0;

  /* dest = src x ret => ~src x dest = ret */
  memmove(&c, src, sizeof(quaternion_t));
  if (invert_quaternion(&c)) return 0;
  return concatenate_quaternions(&c, dest);
}

/* set the (tx, ty, tz) part of a 4x4 matrix with the values from v */
int set_matrix_translation(matrix_t *m, vector_t *v) {
  if (! m) return 1;
  if (! v) return 2;

  M[TX] = v->x;
  M[TY] = v->y;
  M[TZ] = v->z;

  return 0;
}

/* read in three floats */
int read_euler(FILE *fd, euler_t *e) {
  if (! e) {
    ERROR("null euler pointer\n");
    return -1;
  }

  if (read_float(fd, &((*e)[0]))) {
    ERROR("can't read X rotation\n");
    return 1;
  }
  if (read_float(fd, &((*e)[1]))) {
    ERROR("can't read Y rotation\n");
    return 2;
  }
  if (read_float(fd, &((*e)[2]))) {
    ERROR("can't read Z rotation\n");
    return 3;
  }

  return 0;
}

/* read in three floats */
int read_vector(FILE *fd, vector_t *v) {
  if (! v) {
    ERROR("null vector pointer\n");
    return -1;
  }

  if (read_float(fd, &v->x)) {
    ERROR("can't read X\n");
    return 1;
  }
  if (read_float(fd, &v->y)) {
    ERROR("can't read Y\n");
    return 2;
  }
  if (read_float(fd, &v->z)) {
    ERROR("can't read Z\n");
    return 3;
  }

  return 0;
}

/* read in four floats */
int read_quaternion(FILE *fd, quaternion_t *q) {
  if (! q) {
    ERROR("null quaternion pointer\n");
    return -1;
  }

  if (read_float(fd, &q->x)) {
    ERROR("can't read X\n");
    return 1;
  }
  if (read_float(fd, &q->y)) {
    ERROR("can't read Y\n");
    return 2;
  }
  if (read_float(fd, &q->z)) {
    ERROR("can't read Z\n");
    return 3;
  }
  if (read_float(fd, &q->w)) {
    ERROR("can't read W\n");
    return 4;
  }

  return 0;
}

/* read a 3x3 matrix as nine floats */
int read_matrix(FILE *fd, matrix_t *m) {
  if (! m) {
    ERROR("null matrix pointer\n");
    return -1;
  }

  /* first column */
  if (read_float(fd, &(M[R0]))) {
    ERROR("can't read matrix 0th element\n");
    return 1;
  }
  if (read_float(fd, &(M[R1]))) {
    ERROR("can't read matrix 1st element\n");
    return 2;
  }
  if (read_float(fd, &(M[R2]))) {
    ERROR("can't read matrix 2nd element\n");
    return 3;
  }
  /* second column */
  if (read_float(fd, &(M[R3]))) {
    ERROR("can't read matrix 3rd element\n");
    return 4;
  }
  if (read_float(fd, &(M[R4]))) {
    ERROR("can't read matrix 4th element\n");
    return 5;
  }
  if (read_float(fd, &(M[R5]))) {
    ERROR("can't read matrix 5th element\n");
    return 6;
  }
  /* third column */
  if (read_float(fd, &(M[R6]))) {
    ERROR("can't read matrix 6th element\n");
    return 7;
  }
  if (read_float(fd, &(M[R7]))) {
    ERROR("can't read matrix 7th element\n");
    return 8;
  }
  if (read_float(fd, &(M[R8]))) {
    ERROR("can't read matrix 8th element\n");
    return 9;
  }

  return 0;
}

/* write three floats */
int write_euler(FILE *fd, euler_t *e) {
  if (! e) {
    ERROR("null euler pointer\n");
    return -1;
  }

  if (write_float(fd, &((*e)[0]))) {
    ERROR("can't write X rotation\n");
    return 1;
  }
  if (write_float(fd, &((*e)[1]))) {
    ERROR("can't write Y rotation\n");
    return 2;
  }
  if (write_float(fd, &((*e)[2]))) {
    ERROR("can't write Z rotation\n");
    return 3;
  }

  return 0;
}

/* write three floats */
int write_vector(FILE *fd, vector_t *v) {
  if (! v) {
    ERROR("null vector pointer\n");
    return -1;
  }

  if (write_float(fd, &v->x)) {
    ERROR("can't write X\n");
    return 1;
  }
  if (write_float(fd, &v->y)) {
    ERROR("can't write Y\n");
    return 2;
  }
  if (write_float(fd, &v->z)) {
    ERROR("can't write Z\n");
    return 3;
  }

  return 0;
}

/* write four floats */
int write_quaternion(FILE *fd, quaternion_t *q) {
  if (! q) {
    ERROR("null quaternion pointer\n");
    return -1;
  }

  if (write_float(fd, &q->x)) {
    ERROR("can't write quaternion X\n");
    return 1;
  }
  if (write_float(fd, &q->y)) {
    ERROR("can't write quaternion Y\n");
    return 2;
  }
  if (write_float(fd, &q->z)) {
    ERROR("can't write quaternion Z\n");
    return 3;
  }
  if (write_float(fd, &q->w)) {
    ERROR("can't write quaternion W\n");
    return 4;
  }

  return 0;
}

/* write a 3x3 matrix as nine floats */
int write_matrix(FILE *fd, matrix_t *m) {
  if (! m) {
    ERROR("null matrix pointer\n");
    return -1;
  }

  /* first column */
  if (write_float(fd, &(M[R0]))) {
    ERROR("can't write matrix 0th element\n");
    return 1;
  }
  if (write_float(fd, &(M[R1]))) {
    ERROR("can't write matrix 1st element\n");
    return 2;
  }
  if (write_float(fd, &(M[R2]))) {
    ERROR("can't write matrix 2nd element\n");
    return 3;
  }
  /* second column */
  if (write_float(fd, &(M[R3]))) {
    ERROR("can't write matrix 3rd element\n");
    return 4;
  }
  if (write_float(fd, &(M[R4]))) {
    ERROR("can't write matrix 4th element\n");
    return 5;
  }
  if (write_float(fd, &(M[R5]))) {
    ERROR("can't write matrix 5th element\n");
    return 6;
  }
  /* third column */
  if (write_float(fd, &(M[R6]))) {
    ERROR("can't write matrix 6th element\n");
    return 7;
  }
  if (write_float(fd, &(M[R7]))) {
    ERROR("can't write matrix 7th element\n");
    return 8;
  }
  if (write_float(fd, &(M[R8]))) {
    ERROR("can't write matrix 8th element\n");
    return 9;
  }

  return 0;
}


